/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <autosprintf.h>
#include <digitalSTROM/ds485-client-interface.h>

#include "dsmapi-wrapper.h"
#include "exceptions.h"
#include "traffic-decode.h"
#include "util.h"
#include "log.h"
#include "gettext_helper.h"

#define DEFAULT_SYNC_CALL_TIMEOUT   5 // seconds

/// \file dsmapi-wrapper.h

using gnu::autosprintf;

namespace dSA
{

dSMApiWrapper::dSMApiWrapper(std::string ds485_daemon_url)
{
    int ret;

    boost::mutex::scoped_lock lock(m_libdsm_mutex);

    m_dsm_api_handle = DsmApiInitialize();
    if (m_dsm_api_handle == NULL)
    {
        throw _Exception("could not initialize dSM API library");
    }

    ret = DsmApiOpen(m_dsm_api_handle, ds485_daemon_url.c_str(), 0);
    if (ret < 0)
    {
        throw _Exception("could not open connection to " + ds485_daemon_url +
                         " (" + util::str(ret) + ")");
    }

    // register callbacks
    DsmApiCallback_t cb;
    cb.function = (void *) dSMApiWrapper::busStateCallback;
    cb.arg = this;
    DsmApiSetBusStateCallback(m_dsm_api_handle, &cb);

    cb.function = (void *) dSMApiWrapper::busChangeCallback;
    cb.arg = this;
    DsmApiSetBusChangeCallback(m_dsm_api_handle, &cb);

    cb.function = (void *) dSMApiWrapper::deviceAddedCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACCESSIBILITY,
                           EVENT_DEVICE_ACCESSIBILITY_ON,
                           &cb);

    cb.function = (void *) dSMApiWrapper::deviceRemovedCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACCESSIBILITY,
                           EVENT_DEVICE_ACCESSIBILITY_OFF,
                           &cb);

    cb.function = (void *) dSMApiWrapper::devicePingCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_TEST_DEVICE_PRESENCE,
                           0,
                           &cb);

    cb.function = (void *) dSMApiWrapper::deviceActionCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACTION,
                           0,
                           &cb);

    cb.function = (void *) dSMApiWrapper::deviceLocalActionCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_LOCAL_ACTION,
                           0,
                           &cb);

    cb.function = (void *) dSMApiWrapper::deviceConfigCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_CONFIG,
                           0,
                           &cb);

    cb.function = (void *) dSMApiWrapper::commUpstreamShortCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_UPSTREAM_SHORT,
                           &cb);

    cb.function = (void *) dSMApiWrapper::commUpstreamLongCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_UPSTREAM_LONG,
                           &cb);

    cb.function = (void *) dSMApiWrapper::commDownstreamShortCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_SHORT,
                           &cb);

    cb.function = (void *) dSMApiWrapper::commDownstreamLong0Callback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_NO_PARAMETER,
                           &cb);

    cb.function = (void *) dSMApiWrapper::commDownstreamLong8Callback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_8BIT_PARAMETER,
                           &cb);

    cb.function = (void *) dSMApiWrapper::commDownstreamLong16Callback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_16BIT_PARAMETER,
                           &cb);

    cb.function = (void *) dSMApiWrapper::deviceStatisticsCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_STATISTICS,
                           0,
                           &cb);

    cb.function = (void *) dSMApiWrapper::deviceTestTransmissionQualityCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_TEST_TRANSMISSION_QUALITY,
                           0,
                           &cb);
}

void dSMApiWrapper::addListener(dSMApiListener* listener)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    m_callback_listeners.push_back(listener);
}

void dSMApiWrapper::removeListener(dSMApiListener* listener)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == listener)
        {
            i = m_callback_listeners.erase(i);
            break;
        }

        i++;
    }
}

void dSMApiWrapper::busStateCallback(void* arg, bus_state_t state)
{
    static_cast<dSMApiWrapper *>(arg)->notifyBusState(state);
}

void dSMApiWrapper::notifyBusState(bus_state_t state)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->busStateCallback(state);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::busChangeCallback(void* arg, dsid_t *id, int flag)
{
    static_cast<dSMApiWrapper *>(arg)->notifyBusChange(id, flag);
}

void dSMApiWrapper::notifyBusChange(dsid_t *id, int flag)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->busChangeCallback(id, flag);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceAddedCallback(uint8_t error_code, void *arg,
                                       dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid,
                                       uint16_t device_id,
                                       uint16_t zone_id,
                                       uint16_t vendor_id,
                                       uint32_t serial_number)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceAdded(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, zone_id, vendor_id,
            serial_number);
}

void dSMApiWrapper::notifyDeviceAdded(uint8_t error_code, dsid_t src_dsm_dsid,
                                     dsid_t dest_dsm_dsid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     uint32_t serial_number)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener *>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceAddedCallback(error_code, src_dsm_dsid,
                                          dest_dsm_dsid, device_id,
                                          zone_id, vendor_id, serial_number);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }

            i++;
        }
    }
}

void dSMApiWrapper::deviceRemovedCallback(uint8_t error_code, void *arg,
                                         dsid_t src_dsm_dsid,
                                         dsid_t dest_dsm_dsid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         uint32_t serial_number)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceRemoved(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, zone_id, vendor_id,
            serial_number);
}

void dSMApiWrapper::notifyDeviceRemoved(uint8_t error_code, dsid_t src_dsm_dsid,
                                        dsid_t dest_dsm_dsid,
                                        uint16_t device_id, uint16_t zone_id,
                                        uint16_t vendor_id,
                                        uint32_t serial_number)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceRemovedCallback(error_code, src_dsm_dsid,
                                            dest_dsm_dsid, device_id, zone_id,
                                            vendor_id, serial_number);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}


void dSMApiWrapper::devicePingCallback(uint8_t error_code, void *arg,
                                         dsid_t src_dsm_dsid,
                                         dsid_t dest_dsm_dsid,
                                         uint16_t device_id,
                                         uint8_t present)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDevicePing(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, present);
}

void dSMApiWrapper::notifyDevicePing(uint8_t error_code, dsid_t src_dsm_dsid,
                                        dsid_t dest_dsm_dsid,
                                        uint16_t device_id, uint8_t present)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->devicePingCallback(error_code, src_dsm_dsid,
                                         dest_dsm_dsid, device_id, present);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceStatisticsCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceStatistics(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, offset, value);
}

void dSMApiWrapper::notifyDeviceStatistics(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceStatisticsCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id, offset, value);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceActionCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceActionCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            zone_id, group_id, button_nr, click_type);
}
void dSMApiWrapper::notifyDeviceActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceActionCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id,
                        zone_id, group_id, button_nr, click_type);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceLocalActionCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceLocalActionCallback(
            error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            zone_id, state);
}
void dSMApiWrapper::notifyDeviceLocalActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceLocalActionCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id,
                        zone_id, state);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceConfigCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceConfigCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            config_class, config_index, value);
}

void dSMApiWrapper::notifyDeviceConfigCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceConfigCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id,
                        config_class, config_index, value);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceTestTransmissionQualityCallback(uint8_t error_code,
        void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    static_cast<dSMApiWrapper *>(arg)->
        notifyDeviceTestTransmissionQualityCallback(error_code,
                src_dsm_dsid, dest_dsm_dsid, device_id,
                downstream_quality, upstream_quality);
}

void dSMApiWrapper::notifyDeviceTestTransmissionQualityCallback(
        uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceTestTransmissionQualityCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id,
                        downstream_quality, upstream_quality);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commUpstreamShortCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags, uint8_t cflags)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommUpstreamShortCallback(
            error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            circuit_id, is_resend, is_sensor, key_type, click_type, quality,
            flags, cflags);
}

void dSMApiWrapper::notifyCommUpstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags, uint8_t cflags)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commUpstreamShortCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id,
                        circuit_id, is_resend, is_sensor, key_type, click_type,
                        quality, flags, cflags);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commUpstreamLongCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommUpstreamLongCallback(
            error_code, src_dsm_dsid, dest_dsm_dsid, device_id,
            circuit_id, mainqualifier, subqualifier, data, quality,
            flags, cflags);
}

void dSMApiWrapper::notifyCommUpstreamLongCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commUpstreamLongCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, device_id,
                        circuit_id, mainqualifier, subqualifier, data,
                        quality, flags, cflags);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamShortCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamShortCallback(
            error_code, src_dsm_dsid, dest_dsm_dsid, command);
}

void dSMApiWrapper::notifyCommDownstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamShortCallback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, command);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong0Callback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong0Callback(
            error_code, src_dsm_dsid, dest_dsm_dsid, command);
}

void dSMApiWrapper::notifyCommDownstreamLong0Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamLong0Callback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, command);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong8Callback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint8_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong8Callback(
            error_code, src_dsm_dsid, dest_dsm_dsid, command, data);
}

void dSMApiWrapper::notifyCommDownstreamLong8Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint8_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamLong8Callback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, command, data);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong16Callback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint16_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong16Callback(
            error_code, src_dsm_dsid, dest_dsm_dsid, command, data);
}

void dSMApiWrapper::notifyCommDownstreamLong16Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint16_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamLong16Callback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, command, data);
            }
            catch (Exception ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}


void dSMApiWrapper::getOwnDSID(dsid_t* dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetOwnDSID(m_dsm_api_handle, dsid);
    checkRetcode(ret);
}

bus_state_t dSMApiWrapper::getBusState()
{
    bus_state_t state;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetBusState(m_dsm_api_handle, &state);
    checkRetcode(ret);
    return state;
}

void dSMApiWrapper::restartRegistration(dsid_t dsm_dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_restart_registration(m_dsm_api_handle, dsm_dsid);
    checkRetcode(ret);
}

void dSMApiWrapper::restartMeter(dsid_t dsm_dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMReboot(m_dsm_api_handle, dsm_dsid);
    checkRetcode(ret);
}

int dSMApiWrapper::getBusMembers(dsid_t *device_list, int device_list_size)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetBusMembers(m_dsm_api_handle, device_list,
                                  DSM_API_MAX_DEVICES);
    if (ret < 0)
    {
        throw _ExceptionWithCode("could not get bus members", ret);
    }

    return ret;
}

void dSMApiWrapper::dSMInfo(dsid_t dsm_dsid, uint32_t *hw_version,
                           uint32_t *arm_sw_version, uint32_t *dsp_sw_version,
                           uint16_t *api_version, dsid_t *dsid, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    /// \todo make sure to always check the handle and connection state
    ret = ::dSMInfo(m_dsm_api_handle, dsm_dsid, hw_version, arm_sw_version,
                    dsp_sw_version, api_version, dsid, name);
    checkRetcode(ret);
}

void dSMApiWrapper::dSMState(dsid_t dsm_dsid, uint8_t *hardware,
                            uint8_t *communicationdsp, uint8_t *temperature,
                            uint8_t *configuration, uint8_t *ds485bus,
                            uint8_t *powerline)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ::dSMState(m_dsm_api_handle, dsm_dsid, hardware, communicationdsp,
                     temperature, configuration, ds485bus, powerline);
    checkRetcode(ret);
}

void dSMApiWrapper::getPropertyFlags(dsid_t dsm_dsid, uint8_t *flags)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_flags(m_dsm_api_handle, dsm_dsid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::setPropertyFlags(dsid_t dsm_dsid, uint8_t flags)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_flags(m_dsm_api_handle, dsm_dsid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::getPropertyTemperatureLimit(dsid_t dsm_dsid,
                                                uint8_t *warnlevel,
                                                uint8_t *errorlevel)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_temperature_limits(m_dsm_api_handle, dsm_dsid,
                                               warnlevel, errorlevel);
    checkRetcode(ret);
}

void dSMApiWrapper::setPropertyTemperatureLimit(dsid_t dsm_dsid,
                                                uint8_t warnlevel,
                                                uint8_t errorlevel)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_temperature_limits(m_dsm_api_handle, dsm_dsid,
                                               warnlevel, errorlevel);
    checkRetcode(ret);
}

void dSMApiWrapper::getDSMtemperature(dsid_t dsm_dsid, uint16_t *temperature)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_get_current_temperature(m_dsm_api_handle, dsm_dsid,
                                                    temperature);
    checkRetcode(ret);
}

void dSMApiWrapper::getApartmentId(dsid_t dsm_dsid, uint8_t *id)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_apartment_id(m_dsm_api_handle, dsm_dsid, id);
    checkRetcode(ret);
}

void dSMApiWrapper::setApartmentId(dsid_t dsm_dsid, uint8_t id)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_apartment_id(m_dsm_api_handle, dsm_dsid, id);
    checkRetcode(ret);
}

void dSMApiWrapper::getStatisticUpstream(dsid_t dsm_dsid, uint8_t index,
        uint8_t *length, uint8_t *statistics)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMStatistics_get_upstream(m_dsm_api_handle, dsm_dsid, index,
            length, statistics);
    checkRetcode(ret);
}

void dSMApiWrapper::resetStatisticUpstream(dsid_t dsm_dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMStatistics_reset_upstream(m_dsm_api_handle, dsm_dsid);
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationId(dsid_t dsm_dsid, uint32_t *hash,
        uint32_t *configcount)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMConfig_get_hash(m_dsm_api_handle, dsm_dsid, hash, configcount);
    checkRetcode(ret);
}


void dSMApiWrapper::getZoneCount(dsid_t dsm_dsid, uint8_t *count)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneCount(m_dsm_api_handle, dsm_dsid, count);
    checkRetcode(ret);
}

void dSMApiWrapper::getZoneInfo(dsid_t dsm_dsid, uint8_t zone_index,
                               uint16_t *zone_id, uint8_t *virtual_zone_id,
                               uint8_t *number_of_groups, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneInfo_by_index(m_dsm_api_handle, dsm_dsid, zone_index, zone_id,
                            virtual_zone_id, number_of_groups, name);
    checkRetcode(ret);
}

void dSMApiWrapper::getActiveDevicesCountInZone(dsid_t dsm_dsid,
                                               uint16_t zone_id,
                                               uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_only_active(m_dsm_api_handle, dsm_dsid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getInactiveDevicesCountInZone(dsid_t dsm_dsid,
                                                  uint16_t zone_id,
                                                  uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_only_inactive(m_dsm_api_handle, dsm_dsid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getDevicesCountInZone(dsid_t dsm_dsid,
                                          uint16_t zone_id,
                                          uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_all(m_dsm_api_handle, dsm_dsid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getActiveDeviceInfoInZone(dsid_t dsm_dsid,
                                             uint16_t req_zone_id,
                                             uint16_t device_index,
                                             uint16_t *device_id,
                                             uint16_t *vendor_id,
                                             uint16_t *product_id,
                                             uint16_t *function_id,
                                             uint16_t *version,
                                             uint16_t *zone_id,
                                             uint8_t *state,
                                             uint8_t *locked,
                                             uint8_t *output_mode,
                                             uint8_t *groups,
                                             uint8_t *name,
                                             uint32_t *serial_number,
                                             uint8_t *lt_mode)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_index_only_active(m_dsm_api_handle, dsm_dsid,
                                          req_zone_id, device_index, device_id,
                                          vendor_id, product_id, function_id,
                                          version, zone_id, state, locked,
                                          output_mode, lt_mode, groups,
                                          name, serial_number);
    checkRetcode(ret);
}

void dSMApiWrapper::getDeviceInfoById(dsid_t dsm_dsid, uint16_t req_device_id,
                                 uint16_t *device_id, uint16_t *vendor_id,
                                 uint16_t *product_id, uint16_t *function_id,
                                 uint16_t *version, uint16_t *zone_id,
                                 uint8_t *state, uint8_t *locked,
                                 uint8_t *output_mode, uint8_t *groups,
                                 uint8_t *name, uint32_t *serial_number,
                                 uint8_t *lt_mode)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_device_id(m_dsm_api_handle, dsm_dsid, req_device_id,
                                  device_id, vendor_id, product_id,
                                  function_id, version, zone_id, state,
                                  locked, output_mode, lt_mode, groups,
                                  name, serial_number);
    checkRetcode(ret);
}

void dSMApiWrapper::getDeviceInfoByIndex(dsid_t dsm_dsid, uint16_t req_zone_id, uint16_t req_index,
                                 uint16_t *device_id, uint16_t *vendor_id,
                                 uint16_t *product_id, uint16_t *function_id,
                                 uint16_t *version, uint16_t *zone_id,
                                 uint8_t *state, uint8_t *locked, uint8_t *output_mode,
                                 uint8_t *groups, uint8_t *name,
                                 uint32_t *serial_number, uint8_t *lt_mode)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_index(m_dsm_api_handle, dsm_dsid, req_zone_id,
                                  req_index,
                                  device_id, vendor_id, product_id,
                                  function_id, version, zone_id, state,
                                  locked, output_mode, lt_mode, groups,
                                  name, serial_number);
    checkRetcode(ret);
}

void dSMApiWrapper::blinkDevice(dsid_t dsm_dsid, uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceActionRequest_action_blink(m_dsm_api_handle, dsm_dsid,
                                               device_id);
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceConfig_get_sync_8(m_dsm_api_handle, dsm_dsid,
            device_id, config_class, config_index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint16_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceConfig_get_sync_16(m_dsm_api_handle, dsm_dsid,
            device_id, config_class, config_index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t offset, uint8_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint16_t data;
    int ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            &data);
    checkRetcode(ret);
    if (offset & 1)
        *value = data >> 8;
    else
        *value = data & 0xff;
}

void dSMApiWrapper::getStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t offset, uint32_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint16_t datal, datah;
    int ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            &datal);
    checkRetcode(ret);
    ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            &datah);
    checkRetcode(ret);
    *value = datah << 16 | datal;
}

void dSMApiWrapper::reqStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t offset)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceStatisticsRequest_get(m_dsm_api_handle, dsm_dsid,
            device_id, offset);
    checkRetcode(ret);
}

void dSMApiWrapper::getSensorValue(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t index, uint16_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceSensor_get_value_sync(m_dsm_api_handle, dsm_dsid,
            device_id, index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getSensorType(dsid_t dsm_dsid, uint16_t device_id,
        uint8_t index, uint8_t *sensor_present, uint8_t *sensor_type,
        uint8_t sensor_last)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceSensor_get_type_sync(m_dsm_api_handle, dsm_dsid,
            device_id, index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            sensor_present, sensor_type, sensor_type);
    checkRetcode(ret);
}

void dSMApiWrapper::testDeviceTransmissionQuality(dsid_t dsm_dsid,
                                                  uint16_t device_id,
                                                  uint8_t *downstreamQ,
                                                  uint16_t *upstreamQ)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = TestTransmissionQuality_get_sync(m_dsm_api_handle, dsm_dsid,
                                               device_id,
                                               DEFAULT_SYNC_CALL_TIMEOUT,
                                               downstreamQ,
                                               upstreamQ);
    checkRetcode(ret);
}

void dSMApiWrapper::reqDeviceTransmissionQuality(dsid_t dsm_dsid,
                                                 uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = TestTransmissionQuality_get (m_dsm_api_handle, dsm_dsid, device_id);
    checkRetcode(ret);
}

void dSMApiWrapper::reqDevicePresenceTest(dsid_t dsm_dsid,
                                          uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
#if DSM_API_VERSION >= 0x107
    int ret = TestDevicePresence(m_dsm_api_handle, dsm_dsid, device_id);
    checkRetcode(ret);
#endif
}

void dSMApiWrapper::setDisplayMode(dsid_t dsm_dsid, uint8_t flags)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_set_MMI_flag(m_dsm_api_handle, dsm_dsid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::getDisplayMode(dsid_t dsm_dsid, uint8_t* flags)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_get_MMI_flag(m_dsm_api_handle, dsm_dsid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::setCommunicationLogging(dsid_t dsm_dsid, uint8_t state)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_set_communication_log_state(m_dsm_api_handle,
                                                       dsm_dsid, state);
    checkRetcode(ret);
}

void dSMApiWrapper::getCommunicationLogging(dsid_t dsm_dsid, uint8_t* state)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_get_communication_log_state(m_dsm_api_handle,
                                                       dsm_dsid, state);
    checkRetcode(ret);
}

void dSMApiWrapper::pingAutomatic(dsid_t dsm_dsid, uint16_t auto_interval_time)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_automatic(m_dsm_api_handle, dsm_dsid,
                                           auto_interval_time);
    checkRetcode(ret);
}

void dSMApiWrapper::pingOneTimeAutomatic(dsid_t dsm_dsid,
                            uint16_t interval_time, uint16_t auto_interval_time)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_one_time_automatic(m_dsm_api_handle, dsm_dsid,
                                            interval_time, auto_interval_time);
    checkRetcode(ret);
}

void dSMApiWrapper::pingOneTimeOff(dsid_t dsm_dsid, uint16_t interval_time)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_one_time_off(m_dsm_api_handle, dsm_dsid,
                                                                interval_time);
    checkRetcode(ret);
}

void dSMApiWrapper::pingOff(dsid_t dsm_dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_off(m_dsm_api_handle, dsm_dsid);
    checkRetcode(ret);
}

void dSMApiWrapper::getPingIntervals(dsid_t dsm_dsid, uint16_t *interval_auto,
        uint16_t *interval_device)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_get_ping_intervals(m_dsm_api_handle, dsm_dsid,
                                                interval_auto, interval_device);
    checkRetcode(ret);
}

void dSMApiWrapper::setPingIntervals(dsid_t dsm_dsid, uint16_t interval_auto,
                                                    uint16_t interval_device)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_set_ping_intervals(m_dsm_api_handle, dsm_dsid,
                                            interval_auto, interval_device);
    checkRetcode(ret);
}

std::string dSMApiWrapper::error2Message(int err)
{
    switch (err)
    {
        // dSM-API Error Codes
        case ERROR_OK:
            return _("OK");
        case ERROR_WRONG_PARAMETER:
            return _("dSM-API: wrong parameter");
        case ERROR_ZONE_NOT_FOUND:
            return _("dSM-API: zone not found");
        case ERROR_DEVICE_NOT_FOUND:
            return _("dSM-API: device not found");
        case ERROR_GROUP_NOT_FOUND:
            return _("dSM-API: group not found");
        case ERROR_ZONE_CAN_NOT_BE_DELETED:
            return _("dSM-API: zone cannot be deleted");
        case ERROR_GROUP_CAN_NOT_BE_DELETED:
            return _("dSM-API: group cannot be deleted");
        case ERROR_DEVICE_CAN_NOT_BE_DELETED:
            return _("dSM-API: device cannot be deleted");
        case ERROR_NO_FURTHER_ZONES:
            return _("dSM-API: no further zones");
        case ERROR_NO_FURTHER_GROUPS:
            return _("dSM-API: no further groups");
        case ERROR_ZONE_ALREADY_EXISTS:
            return _("dSM-API: zone already exists");
        case ERROR_GROUP_ALREADY_EXISTS:
            return _("dSM-API: group already exists");
        case ERROR_ZONE_NOT_EMPTY:
            return _("dSM-API zone not empty");
        case ERROR_TIMEOUT:
            return _("dSM-API: command timeout");
        case ERROR_WRONG_SIZE:
            return _("dSM-API: wrong size");
        case ERROR_WRONG_MSGID:
            return _("dSM-API: wrong message id");
        case ERROR_WRONG_MODIFIER:
            return _("dSM-API: wrong modifier");
        case ERROR_WRONG_PACKET_NR:
            return _("dSM-API: wrong packet number");
        case ERROR_WRONG_IMAGE_SIZE:
            return _("dSM-API: wrong image size");
        case ERROR_NO_IMAGE_TRANSFER_ACTIVE:
            return _("dSM-API: no image transfer active");
        case ERROR_IMAGE_INVALID:
            return _("dSM-API: image is invalid");
        case ERROR_NO_CONFIG:
            return _("dSM-API: no configuration present");
        case ERROR_REQUEST_CAN_NOT_BE_EXECUTED:
            return _("dSM-API: request cant be executed");
        case ERROR_OUT_OF_RESOURCES:
            return _("dSM-API: out of ressources");
#if DSM_API_VERSION >= 0x106
        case ERROR_PROGRAMMING_MODE_IS_DISABLED:
            return _("dSM-API: programming mode is disabled");
#endif

        // dS485-client/ds485d error codes
        case ERROR_RESPONSE_TIMEOUT:
            return _("dS485: command timeout");
        case NOTICE_CALL_TO_BROADCAST:
            return _("dS485: broadcast call");
        case ERROR_INVALID_HANDLE:
            return _("dS485: invalid socket handle");
        case ERROR_INVALID_CONNSPEC:
            return _("dS485: invalid connection data");
        case ERROR_THREAD_START_FAILED:
            return _("dS485: thread startup failed");
        case ERROR_NO_PORT_FOUND:
            return _("dS485: network startup failed: no port found");
        case ERROR_INVALID_ADDRESS:
            return _("dS485: network startup failed: invalid address");
        case ERROR_CREATE_SOCKET:
            return _("dS485: network startup failed: no socket");
        case ERROR_SOCKET_OPTIONS:
            return _("dS485: network startup failed: setting socket options");
        case ERROR_SOCKET_CONNECT:
            return _("dS485: network connect failed");
        case ERROR_SOCKET_FULL:
            return _("dS485: network failure: socket full");
        case ERROR_SOCKET_FAILED:
            return _("dS485: network failure: socket write failed");
        case ERROR_SOCKET_VANISHED:
            return _("dS485: network failure: remote closed connection");
        case ERROR_SOCKET_WOULD_BLOCK:
            return _("dS485: network failure: socket blocked");
        case ERROR_SOCKET_UNKNOWN:
            return _("dS485: general network failure");
        case ERROR_INVALID_FD:
            return _("dS485: invalid descriptor");
        case ERROR_INVALID_PARAMETER:
            return _("dS485: invalid parameter");

        default:
            return autosprintf("%s %d", _("Error code"), err);
    }
}

void dSMApiWrapper::checkRetcode(int ret)
{
    if (ret != 0)
    {
        throw _ExceptionWithCode(error2Message(ret) + ": " + util::str(ret),
                                 ret);
    }
}

dSMApiWrapper::~dSMApiWrapper()
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    if (m_dsm_api_handle)
    {
        DsmApiClose(m_dsm_api_handle);
        DsmApiCleanup(m_dsm_api_handle);
    }
}

dSMApiSelectiveListener::CBData::CBData(
        dSMApiSelectiveListener::cb_selection type) : m_data_type(type)
{
    m_timestamp = time(NULL);
}

dSMApiSelectiveListener::CBData::CBData()
{
    m_timestamp = time(NULL);
}

dSMApiSelectiveListener::cb_selection
    dSMApiSelectiveListener::CBData::getType() const
{
    return m_data_type;
}

time_t dSMApiSelectiveListener::CBData::getTimestamp() const
{
    return m_timestamp;
}

dSMApiSelectiveListener::CBCommonData::CBCommonData(
    dSMApiSelectiveListener::cb_selection type, uint8_t error_code,
    dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid) :
    CBData(type), m_error_code(error_code), m_src_dsm_dsid(src_dsm_dsid), m_dest_dsm_dsid(dest_dsm_dsid)
{}

uint8_t dSMApiSelectiveListener::CBCommonData::getErrorCode() const
{
    return m_error_code;
}

dsid_t dSMApiSelectiveListener::CBCommonData::getDSMSourceDSID() const
{
    return m_src_dsm_dsid;
}

dsid_t dSMApiSelectiveListener::CBCommonData::getDSMDestinationDSID() const
{
    return m_dest_dsm_dsid;
}

dSMApiSelectiveListener::BasicDeviceData::BasicDeviceData(
    dSMApiSelectiveListener::cb_selection type,
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id) :
    CBCommonData(type, error_code, src_dsm_dsid, dest_dsm_dsid),
    m_device_id(device_id)
{}

uint16_t dSMApiSelectiveListener::BasicDeviceData::getDeviceID() const
{
    return m_device_id;
}

dSMApiSelectiveListener::DeviceData::DeviceData(
    dSMApiSelectiveListener::cb_selection type,
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint16_t zone_id, uint16_t vendor_id,
    uint32_t serial_number) :
        BasicDeviceData(type, error_code, src_dsm_dsid, dest_dsm_dsid,
        device_id), m_zone_id(zone_id), m_vendor_id(vendor_id),
        m_serial_number(serial_number)
{}

uint16_t dSMApiSelectiveListener::DeviceData::getZoneID() const
{
    return m_zone_id;
}

uint16_t dSMApiSelectiveListener::DeviceData::getVendorID() const
{
    return m_vendor_id;
}

uint16_t dSMApiSelectiveListener::DeviceData::getSerialNumber() const
{
    return m_serial_number;
}

dSMApiSelectiveListener::DeviceAddedData::DeviceAddedData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint16_t zone_id, uint16_t vendor_id,
    uint32_t serial_number) :
        DeviceData(dSMApiSelectiveListener::DeviceAdded, error_code,
        src_dsm_dsid, dest_dsm_dsid, device_id, zone_id, vendor_id,
        serial_number)
{}

dSMApiSelectiveListener::DeviceRemovedData::DeviceRemovedData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint16_t zone_id, uint16_t vendor_id,
    uint32_t serial_number) :
        DeviceData(dSMApiSelectiveListener::DeviceRemoved, error_code,
        src_dsm_dsid, dest_dsm_dsid, device_id, zone_id, vendor_id,
        serial_number)
{}

dSMApiSelectiveListener::DeviceTestQualityData::DeviceTestQualityData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint8_t downstream, uint16_t upstream) :
        BasicDeviceData(dSMApiSelectiveListener::DeviceTestTransmissionQuality,
                error_code, src_dsm_dsid, dest_dsm_dsid, device_id),
        m_down(downstream), m_up(upstream)
{}

uint8_t dSMApiSelectiveListener::DeviceTestQualityData::getDownstream() const
{
    return m_down;
}

uint16_t dSMApiSelectiveListener::DeviceTestQualityData::getUpstream() const
{
    return m_up;
}

dSMApiSelectiveListener::DevicePingData::DevicePingData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint8_t present) :
        BasicDeviceData(dSMApiSelectiveListener::DevicePing,
                error_code, src_dsm_dsid, dest_dsm_dsid, device_id),
        m_present(present)
{}

uint8_t dSMApiSelectiveListener::DevicePingData::getPresent() const
{
    return m_present;
}

dSMApiSelectiveListener::DeviceStatisticData::DeviceStatisticData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint8_t offset, uint16_t value) :
        BasicDeviceData(dSMApiSelectiveListener::DeviceStatistics,
                error_code, src_dsm_dsid, dest_dsm_dsid, device_id),
        m_offset(offset), m_value(value)
{}

uint8_t dSMApiSelectiveListener::DeviceStatisticData::getOffset() const
{
    return m_offset;
}

uint16_t dSMApiSelectiveListener::DeviceStatisticData::getValue() const
{
    return m_value;
}

dSMApiSelectiveListener::CommData::CommData(cb_selection type,
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint8_t circuit_id, uint8_t quality,
    uint8_t flags, uint8_t cflags) :
    BasicDeviceData(type, error_code, src_dsm_dsid, dest_dsm_dsid, device_id),
    m_circuit_id(circuit_id), m_quality(quality),
    m_flags(flags), m_cflags(cflags)
{}

uint8_t dSMApiSelectiveListener::CommData::getCircuitID() const
{
    return m_circuit_id;
}

uint8_t dSMApiSelectiveListener::CommData::getQuality() const
{
    return m_quality;
}

uint8_t dSMApiSelectiveListener::CommData::getFlags() const
{
    return m_flags;
}

uint8_t dSMApiSelectiveListener::CommData::getCrosstalkFlags() const
{
    return m_cflags;
}

dSMApiSelectiveListener::CommUpstreamShortData::CommUpstreamShortData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint8_t circuit_id, uint8_t is_resend,
    uint8_t is_sensor, uint8_t key_type, uint8_t click_type, uint8_t quality,
    uint8_t flags, uint8_t cflags) :
        CommData(dSMApiSelectiveListener::CommUpstreamShort, error_code,
        src_dsm_dsid, dest_dsm_dsid, device_id, circuit_id, quality,
        flags, cflags),
        m_is_resend(is_resend), m_is_sensor(is_sensor),
        m_key_type(key_type), m_click_type(click_type)
{}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::isResend() const
{
    return m_is_resend;
}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::isSensor() const
{
    return m_is_sensor;
}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::getKeyType() const
{
    return m_key_type;
}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::getClickType() const
{
    return m_click_type;
}

dSMApiSelectiveListener::CommUpstreamLongData::CommUpstreamLongData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint16_t device_id, uint8_t circuit_id, uint8_t mainqualifier,
    uint8_t subqualifier, uint32_t data, uint8_t quality,
    uint8_t flags, uint8_t cflags) :
        CommData(dSMApiSelectiveListener::CommUpstreamLong, error_code,
        src_dsm_dsid, dest_dsm_dsid, device_id, circuit_id, quality,
        flags, cflags),
        m_mainqualifier(mainqualifier), m_subqualifier(subqualifier),
        m_data(data)
{}

uint8_t dSMApiSelectiveListener::CommUpstreamLongData::getMainQualifier() const
{
    return m_mainqualifier;
}

uint8_t dSMApiSelectiveListener::CommUpstreamLongData::getSubQualifier() const
{
    return m_subqualifier;
}

uint32_t dSMApiSelectiveListener::CommUpstreamLongData::getData() const
{
    return m_data;
}

dSMApiSelectiveListener::CommDownstreamData::CommDownstreamData(
    dSMApiSelectiveListener::cb_selection type, uint8_t error_code,
    dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command) :
    CBCommonData(type, error_code, src_dsm_dsid, dest_dsm_dsid),
    m_command(command)
{}

uint8_t dSMApiSelectiveListener::CommDownstreamData::getCommand() const
{
    return m_command;
}

dSMApiSelectiveListener::CommDownstreamShortData::CommDownstreamShortData(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint8_t command) :
        CommDownstreamData(dSMApiSelectiveListener::CommDownstreamShort,
        error_code, src_dsm_dsid, dest_dsm_dsid, command)
{}

dSMApiSelectiveListener::CommDownstreamLong0Data::CommDownstreamLong0Data(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint8_t command) :
        CommDownstreamData(dSMApiSelectiveListener::CommDownstreamLong0,
        error_code, src_dsm_dsid, dest_dsm_dsid, command)
{}

dSMApiSelectiveListener::CommDownstreamLong8Data::CommDownstreamLong8Data(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint8_t command, uint8_t data) :
        CommDownstreamData(dSMApiSelectiveListener::CommDownstreamLong8,
        error_code, src_dsm_dsid, dest_dsm_dsid, command), m_data(data)
{}

uint8_t dSMApiSelectiveListener::CommDownstreamLong8Data::getData() const
{
    return m_data;
}

dSMApiSelectiveListener::CommDownstreamLong16Data::CommDownstreamLong16Data(
    uint8_t error_code, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
    uint8_t command, uint16_t data) :
        CommDownstreamData(dSMApiSelectiveListener::CommDownstreamLong16,
        error_code, src_dsm_dsid, dest_dsm_dsid, command), m_data(data)
{}

uint16_t dSMApiSelectiveListener::CommDownstreamLong16Data::getData() const
{
    return m_data;
}

bool dSMApiSelectiveListener::checkCB(
        dSMApiSelectiveListener::cb_selection sel)
{
    boost::recursive_mutex::scoped_lock lock(m_sel_access_mutex);
    if (m_selection[sel])
    {
        return true;
    }

    return false;
}

dSMApiSelectiveListener::dSMApiSelectiveListener()
{
    m_selection.resize(CB_MAX);
    for (size_t i = 0; i < m_selection.size(); i++)
    {
        m_selection[i] = false;
    }
}

void dSMApiSelectiveListener::select(
        dSMApiSelectiveListener::cb_selection cb, bool on)
{
    boost::recursive_mutex::scoped_lock lock(m_sel_access_mutex);
    m_selection[cb] = on;
}

dSMApiSelectiveListener::~dSMApiSelectiveListener() {}

void dSMApiSelectiveListener::busStateCallback(bus_state_t state)
{
    if (!checkCB(BusState))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::busChangeCallback(dsid_t *id, int flag)
{
    if (!checkCB(BusChange))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceAddedCallback(uint8_t error_code,
                                    dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
                                    uint16_t device_id, uint16_t zone_id,
                                    uint16_t vendor_id, uint32_t serial_number)
{
    if (!checkCB(DeviceAdded))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceAddedData> d(
                new dSMApiSelectiveListener::DeviceAddedData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            device_id, zone_id, vendor_id,
                                            serial_number));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::deviceAddedCallback caught exception";
    }
}

void dSMApiSelectiveListener::deviceRemovedCallback(uint8_t error_code,
                                        dsid_t src_dsm_dsid,
                                        dsid_t dest_dsm_dsid,
                                        uint16_t device_id,
                                        uint16_t zone_id, uint16_t vendor_id,
                                        uint32_t serial_number)
{
    if (!checkCB(DeviceRemoved))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceRemovedData> d(
            new dSMApiSelectiveListener::DeviceRemovedData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            device_id, zone_id, vendor_id,
                                            serial_number));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::deviceRemovedCallback caught exception";
    }
}

void dSMApiSelectiveListener::devicePingCallback(uint8_t error_code,
                                        dsid_t src_dsm_dsid,
                                        dsid_t dest_dsm_dsid,
                                        uint16_t device_id,
                                        uint8_t present)
{
    if (!checkCB(DevicePing))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DevicePingData> d(
            new dSMApiSelectiveListener::DevicePingData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            device_id, present));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::devicePingCallback caught exception";
    }
}

void dSMApiSelectiveListener::deviceStatisticsCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value)
{
    if (!checkCB(DeviceStatistics))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceStatisticData> d(
            new dSMApiSelectiveListener::DeviceStatisticData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid, device_id,
                                            offset, value));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
}

void dSMApiSelectiveListener::deviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type)
{
    if (!checkCB(DeviceAction))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state)
{
    if (!checkCB(DeviceLocalAction))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value)
{
    if (!checkCB(DeviceConfig))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceTestTransmissionQualityCallback(
            uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality)
{
    if (!checkCB(DeviceTestTransmissionQuality))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceTestQualityData> d(
            new dSMApiSelectiveListener::DeviceTestQualityData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid, device_id,
                                            downstream_quality, upstream_quality));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
}

void dSMApiSelectiveListener::commUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type, uint8_t quality,
            uint8_t flags, uint8_t cflags)
{
    if (!checkCB(CommUpstreamShort))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommUpstreamShortData> d(
            new dSMApiSelectiveListener::CommUpstreamShortData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            device_id, circuit_id, is_resend,
                                            is_sensor, key_type, click_type,
                                            quality, flags, cflags));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commUpstreamShortCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
    if (!checkCB(CommUpstreamLong))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommUpstreamLongData> d(
            new dSMApiSelectiveListener::CommUpstreamLongData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            device_id, circuit_id,
                                            mainqualifier, subqualifier,
                                            data, quality, flags, cflags));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commUpstreamLongCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    if (!checkCB(CommDownstreamShort))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamShortData> d(
            new dSMApiSelectiveListener::CommDownstreamShortData(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            command));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamShortCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    if (!checkCB(CommDownstreamLong0))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamLong0Data> d(
            new dSMApiSelectiveListener::CommDownstreamLong0Data(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            command));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamLong0Callback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint8_t data)
{
    if (!checkCB(CommDownstreamLong8))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamLong8Data> d(
            new dSMApiSelectiveListener::CommDownstreamLong8Data(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            command, data));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamLong8Callback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint16_t data)
{
    if (!checkCB(CommDownstreamLong16))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamLong16Data> d(
            new dSMApiSelectiveListener::CommDownstreamLong16Data(error_code,
                                            src_dsm_dsid, dest_dsm_dsid,
                                            command, data));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamLong16Callback"
            << " caught exception";
    }
}

}; // namespace

